#include <stddef.h>
#include <string.h>
#include <piece.h>

#include "class.h"
#include "util.h"
#include "alloc.h"
#include "loader.h"
#include "error.h"

#define CLASS_POOL_SIZE 23

static struct jclass *class_bucket[CLASS_POOL_SIZE];

void register_class( struct jclass *cls )
{
	unsigned long h = hash_string(cls->classname,strlen(cls->classname))%CLASS_POOL_SIZE;
	struct jclass **bucket = class_bucket+h;

	while ( *bucket != NULL ) {
		if ( strcmp( (*bucket)->classname,cls->classname)==0 )
			return;
		
		bucket = &((*bucket)->next);
	}
	cls->next = NULL;
	*bucket = cls;
}

struct jclass *get_class( const char *name ) {
	unsigned long h = hash_string(name,strlen(name))%CLASS_POOL_SIZE;
	struct jclass *cls = class_bucket[h];

	while ( cls != NULL ) {
		if ( strcmp( cls->classname, name ) == 0 )
			return cls;
		
		cls = cls->next;
	}

	/* $B8+$D$+$i$J$+$C$?$N$G?7$7$/%m!<%I$9$k(B */
	cls = load_new_class( name );
	if ( cls != NULL ) {
		register_class( cls );
		return cls;
	}

	{
		char error[256];
		sprintf( error, "class %s can't load.", name );
		fetal_error( error );
	}
	return NULL;
}

int compute_argnum( const char *desc )
{
	int i=0;
	const char *p = desc+1;

	while ( *p != ')' ) {
		switch ( *p ) {
		case 'L':
		case 'I':
		case 'B':
		case 'C':
		case 'Z':
			i++;
			break;

		default:
			;
		}
		p++;
	}
	return i;
}

struct method_header *get_method( struct jclass *c, const char *name, const char *desc, int mode )
{
	int i;
	int num;

	if ( mode==GET_METHOD_VIRTUAL ) {
		num = c->vtbl_size;
	} else {
		num = c->imeth_cnt;
	}

	for ( i=0; i<num; i++ ) {
		struct method_header *m;
		if ( mode == GET_METHOD_VIRTUAL )
			m = c->vtbl[i];
		else
			m = c->instance_method+i;
		if ( (strcmp(m->name,name)==0) &&
		     (strcmp(m->desc,desc)==0) ) 
			return m;
	}

	{
		char errormes[256];
		sprintf( errormes, "not found method, %s.%s\n", name ,desc );
		fetal_error(errormes);
	}
	return NULL;		/* $B$3$3$^$GMh$J$$(B */
}

/* $B%9%?%F%#%C%/%a%=%C%I$r?F%/%i%9$^$GAL$C$FD4$Y$kI,MW$C$F$"$k(B? */
struct method_header *get_static_method( struct jclass *c, const char*name, const char*desc )
{
	int i;

	for ( i=0; i<c->smeth_cnt; i++ ) {
		struct method_header *m = &(c->static_method[i]);
		if ( (strcmp(m->name,name)==0) &&
		     (strcmp(m->desc,desc)==0) ) 
			return m;
	}

	{
		char errmes[256];
		sprintf( errmes, "not found method, %s.%s", name ,desc );
		fetal_error(errmes);
	}
	return NULL;
}

#include "jobject.h"
#include "jtestobject.h"
#include "jstring.h"
#include "stringbuffer.h"
#include "piece.h"
#include "piece_bmp.h"

/* $B%M%$%F%#%V%/%i%9$NEPO?(B */
void init_class_table()
{
	register_class( new_jobject_class() );

#ifndef NATIVE
	register_class( new_testobject_class() );
#endif	

	register_class( new_jstring_class() );
	register_class( new_stringbuffer_class() );

#ifdef USE_PIECE	
	register_class( new_piece_class() );
	register_class( init_piece_bmp_class() );
#endif	

}

		
