#include "class.h"
#include "pool.h"
#include "readbuf.h"
#include "alloc.h"
#include <string.h>
#include "resolve.h"
#include "error.h"

/* $B$$$i$J$$(Battribute$B$O<N$F$k(B */
static void discard_attribute( READ_BUF *buf )
{
	int length;
	int i;
	read_u2(buf);
	length = read_u4(buf);

	for ( i=0; i<length; i++ )
		read_u1(buf);
}


static char *utf8(READ_BUF *rbuf)
{
	int i;
	u2 length = read_u2(rbuf);
	char *buf = xmalloc(length+1);
	for ( i=0; i<length; i++ )
		buf[i]=read_u1(rbuf);
	buf[i]='\0';
	return buf;
}

static struct exception_table *read_except_table( READ_BUF *buf, struct const_pool *consts,
						  int *length, u1 *code_base )
{
	u2 len = read_u2(buf);
	int i;
	struct exception_table *tbl;
	if ( len == 0 )
		return NULL;

	*length = len;
	tbl = xmalloc( sizeof(struct exception_table)*len );

	for ( i=0; i<len; i++ ) {
		u2 catch_type;
		tbl[i].start_pc = read_u2(buf)+code_base;
		tbl[i].end_pc = read_u2(buf)+code_base;
		tbl[i].handler_pc = read_u2(buf)+code_base;

		catch_type = read_u2(buf);
		tbl[i].catch_type = get_class((char*)consts[consts[catch_type].val1].val1);
	}

	return tbl;
}

static void read_methatt( READ_BUF *buf, struct method_header *meth, struct jclass *cls )
{
	const char *attname = STRING_CONST( cls, read_u2(buf) );
	u4 length = read_u4(buf);
	if ( strcmp( attname, "Code" ) == 0 ) {
		char *code;
		u4 code_length;
		u2 attributes_count;
		int except_len;
		int i;

		read_u2(buf);	/* max stack */
		meth->max_local = read_u2(buf);
		code_length = read_u4(buf);
		
		code = xmalloc(code_length);
		read_buf( buf, code_length, code );
		meth->code = code;

		meth->except_tbl = read_except_table( buf, cls->constants, &except_len, code );
		meth->except_tbl_len = except_len;

		attributes_count = read_u2(buf);

		for ( i=0; i< attributes_count; i++ )
			discard_attribute(buf);

	} else {
		int i;
		for ( i=0; i<length; i++ ) {
			read_u1(buf);
		}
	}	
}

void load_pool( struct const_pool *ptr, int count, READ_BUF *buf )
{
	int i;
	for ( i=1; i<count; i++ ) {
		u1 tag = read_u1(buf);
		ptr[i].type = tag;
		
		switch ( tag ) {
		case tag_UTF8: {
			char *str = utf8(buf);
			ptr[i].val1 = (u4)str;
			break;
		}

		case tag_INT:
			ptr[i].val1 = read_u4(buf);
			break;
			
		case tag_CLASS:
		case tag_STRING:
			ptr[i].val1 = read_u2(buf);
			break;

		case tag_FIELDREF: /* val1 ->class | val2 ->name */
		case tag_METHODREF:
		case tag_INTFCMETHREF:
		case tag_NAMEANDTYPE: /* val1 ->name | val2 ->type */
			ptr[i].val1 = read_u2(buf);
			ptr[i].val2 = read_u2(buf);
			break;

		default:
			fetal_error("unrecognize constant pool tag.");
			break;
		}
	}
}

struct first_path
{
	int smeth_cnt;		/* $B%9%?%F%#%C%/%a%=%C%I$N?t(B */
	int imeth_cnt;		/*  */
	
	int static_fieldnum;	
	int instance_fieldnum;
};
		
/* $B>e$N9=B$BN;2>H!%%a%=%C%I$H$+$N?t$r?t$($k(B */
static void first_path( struct first_path *fp, READ_BUF *buf, struct jclass *cls )
{
	unsigned long pos = buf_tell(buf);

	int ifnum=0, sfnum=0;
	int smeth_cnt=0, imeth_cnt=0;

	int field_cnt = read_u2(buf);
	int meth_cnt;
	int i;

	for ( i=0; i<field_cnt; i++ ) {
		int attcnt;
		int j;
		int access = read_u2(buf);
		if ( access&ACC_STATIC )
			sfnum++;
		else
			ifnum++;
		read_u4(buf);
		attcnt = read_u2(buf);
		for ( j=0; j<attcnt; j++ ) 
			discard_attribute(buf);
	}

	meth_cnt = read_u2(buf);

	for ( i=0; i<meth_cnt; i++ ) {
		int attcnt;
		int j;
		int acc = read_u2(buf);

		read_u4(buf);	/* access_flag, name_index */

		if ( acc&ACC_STATIC ) {
			smeth_cnt++;
		} else {
			imeth_cnt++;
		}
		
		attcnt = read_u2(buf);
		for ( j=0; j<attcnt; j++ )
			discard_attribute(buf);
	}

	buf_setpos(buf,pos);
	fp->smeth_cnt=smeth_cnt;
	fp->imeth_cnt=imeth_cnt;
	fp->static_fieldnum=sfnum;
	fp->instance_fieldnum=ifnum;
}

static void load_field( struct jclass *cls, struct first_path *_1path, READ_BUF *buf )
{
	int field_cnt = read_u2(buf);
	int i;
	int field_offset;
	int static_offset=0;

	struct field_bucket *instance_p=NULL;
	struct field_bucket *static_p=NULL;

	if ( _1path->static_fieldnum!=0 )
		static_p = cls->static_field = xmalloc(sizeof(struct field_bucket)*_1path->static_fieldnum);
	if ( _1path->instance_fieldnum!=0 )
		instance_p = xmalloc( sizeof(struct field_bucket)*_1path->instance_fieldnum);
	cls->instance_field = instance_p;

	cls->static_field_num = _1path->static_fieldnum;
	cls->instance_field_num = _1path->instance_fieldnum;

	if ( _1path->static_fieldnum != 0 )
		cls->class_field_val = xcalloc( _1path->static_fieldnum, sizeof(jvalue) );
	cls->field_size = cls->parent->field_size+cls->instance_field_num;
	field_offset = cls->parent->field_size;

	for ( i=0; i<field_cnt; i++ ) {
		int access=read_u2(buf);
		int attcnt,j;

		if ( access&ACC_STATIC ) {
			static_p->name=STRING_CONST(cls,read_u2(buf));
			static_p->offset=static_offset++;
			static_p++;
		} else {
			instance_p->name = STRING_CONST(cls,read_u2(buf));
			instance_p->offset = field_offset++;
			instance_p++;
		}
		read_u2(buf);	/* descriptor */
		
		attcnt = read_u2(buf);

		/* constant value $B$O(B... ? */
		for ( j=0; j<attcnt; j++ )
			discard_attribute(buf);
	}


}

#define METHOD_EQUAL(m1,m2) \
((strcmp((m1).name,(m2).name) == 0) &&\
 (strcmp((m1).desc,(m2).desc) == 0))

/* $B%*!<%P!<%i%$%I$7$F$$$k%a%=%C%I$N?t$r?t$($k$N$H!"(B
 * $B%*!<%P!<%i%$%I$7$F$?$i!"$=$l$N(Bvtable$B%*%U%;%C%H$r?F%/%i%9$N$=$l$HF1$8$K$9$k(B
 */
static int count_override( struct jclass *cls )
{
	int i;
	const struct jclass * const parent = cls->parent;
	int count=0;

	for ( i=0; i<cls->imeth_cnt; i++ ) {
		int j;
		for ( j=0; j<parent->imeth_cnt; j++ ) {
			struct method_header *m = &parent->instance_method[j];
			if ( METHOD_EQUAL(cls->instance_method[i], *m) ) {
				count++;
				cls->instance_method[i].offset = m->offset;
			}
		}
	}

	return count;
}

static void make_vtbl( struct jclass *cls )
{
	const int override_cnt = count_override(cls);
	int offset_cnt;
	int i;
	cls->vtbl_size += cls->parent->vtbl_size - override_cnt;
	cls->vtbl = xmalloc( sizeof(struct method_header*) * (cls->vtbl_size) );

	memcpy( cls->vtbl, cls->parent->vtbl, sizeof(struct method_header*)*cls->parent->vtbl_size );
	offset_cnt = cls->parent->vtbl_size;

	for ( i=0; i<cls->imeth_cnt; i++ ) {
		struct method_header *m = &cls->instance_method[i];
		if ( m->offset < 0 ) { /* $B%*!<%P!<%i%$%I$7$F$J$$%a%=%C%I(B */
			cls->vtbl[offset_cnt] = m;
			m->offset = offset_cnt++;
		} else {
			cls->vtbl[m->offset] = m;
		}
	}
}

static void load_method( struct jclass *cls, struct first_path *_1path, READ_BUF *buf )
{
	const int count = read_u2(buf);
	int i;

	struct method_header *methods = xmalloc( sizeof(struct method_header) *
						 (_1path->smeth_cnt+_1path->imeth_cnt) );

	struct method_header *s_method = methods;
	struct method_header *i_method = methods+_1path->smeth_cnt;

	/*
	  struct method_header *s_method = xmalloc( sizeof(struct method_header) * _1path->smeth_cnt );
	  struct method_header *i_method = xmalloc( sizeof(struct method_header) * _1path->imeth_cnt);
	*/

	cls->static_method = s_method;
	cls->instance_method = i_method;

	cls->vtbl_size = _1path->imeth_cnt;

	cls->imeth_cnt = _1path->imeth_cnt;
	cls->smeth_cnt = _1path->smeth_cnt;

	for ( i=0; i<count; i++ ) {
		struct method_header *now_method;
		const int access = read_u2(buf);
		const char *name = STRING_CONST(cls,read_u2(buf));
		const char *desc = STRING_CONST(cls,read_u2(buf));
		int attcnt;
		int argnum = compute_argnum(desc);

		int j;

		if ( access&ACC_STATIC ) {
			now_method = s_method;
			s_method++;
		} else {
			now_method = i_method;
			i_method++;
			argnum++; /* this$B$,0z?t$K$"$k(B */
		}

		now_method->name = name;
		now_method->desc = desc;
		now_method->argnum = argnum;
		now_method->ofclass = cls;
		now_method->offset = -1;
		now_method->is_native = 0;

		attcnt = read_u2(buf);
		for ( j=0; j<attcnt; j++ )
			read_methatt( buf, now_method, cls );
	}

	make_vtbl( cls );
}

/* jclass $B$r:n$k(B */
struct jclass *load_class( READ_BUF *buf )
{
	int pool_cnt;
	int interface_cnt;

	struct jclass *cls = xmalloc( sizeof(struct jclass) );

	struct first_path _1path;
	
	if ( read_u4(buf) != 0xCAFEBABE ) { /* magic number */
		fetal_error("can't load class file.");
		return NULL;
	}
	
	cls = xmalloc( sizeof(struct jclass) );

	read_u4(buf);		/* major version $B$H(B minor version */

	cls->vtbl_size=0;
	
	pool_cnt = read_u2(buf);	/* constant pool count */
	cls->pool_num = pool_cnt;
	cls->constants = xmalloc( sizeof(struct const_pool)*pool_cnt );
	load_pool( cls->constants, pool_cnt, buf );

	/* general information */
	read_u2(buf);		/* access_flag */

	cls->classname = STRING_CONST(cls,cls->constants[read_u2(buf)].val1);
	register_class( cls );
	cls->parent = get_class( STRING_CONST(cls, cls->constants[read_u2(buf)].val1) );
	if ( cls->parent == NULL ) {
		fetal_error("parent $B$,8+$D$+$i$X$s(B.");
	}

	/* interface */
	interface_cnt = read_u2(buf);
	if ( interface_cnt != 0 ) {
		fetal_error("can't use interface.");
	}

	first_path( &_1path, buf, cls );
	load_field( cls, &_1path, buf );

	load_method( cls, &_1path, buf );

	resolve_constant( cls );

	return cls;

//error:
	// free_class(cls);
	return NULL;
}

struct jclass *load_new_class( const char *name )
{
	char *fname=NULL;
	READ_BUF *fp=NULL;
	struct jclass *ret=NULL;
	fname = xmalloc(strlen(name)+6+1); /* ".class"=5 */
	strcpy( fname, name );

	strcat( fname, ".class" );

	fp = readbuf_open( fname );
	if ( fp == NULL )
		goto error;

	ret = load_class( fp );
	if ( ret == NULL )
		goto error;

	goto exit;

error:
	if ( fname ) xfree(fname);
	if ( ret ) xfree(ret);

exit:
	if ( fp ) readbuf_close(fp);
	return ret;
}

	
