#include "class.h"
#include "pool.h"
#include "readbuf.h"
#include "alloc.h"
#include <string.h>
#include "resolve.h"
#include "error.h"

static char *utf8(READ_BUF *rbuf)
{
	int i;
	u2 length = read_u2(rbuf);
	char *buf = xmalloc(length+1);
	for ( i=0; i<length; i++ )
		buf[i]=read_u1(rbuf);
	buf[i]='\0';
	return buf;
}

static void read_methatt( READ_BUF *buf, struct method_list *meth, struct jclass *cls )
{
	const char *attname = STRING_CONST( cls, read_u2(buf) );
	u4 length = read_u4(buf);
	if ( strcmp( attname, "Code" ) == 0 ) {
		char *code;
		u4 code_length;
		u2 exp_tbl_length;
		int i;

		read_u2(buf);	/* max stack */
		meth->max_local = read_u2(buf);
		code_length = read_u4(buf);
		
		code = xmalloc(code_length);
		read_buf( buf, code_length, code );
		meth->code = code;

		exp_tbl_length = read_u2(buf);

		for ( i=0; i<exp_tbl_length; i++ ) {
			char tmp[8];
			read_buf( buf, 8, tmp );
		}

		/* exception_table $B$H(B attributes
		 * attributes$B$O%G%P%C%0;~$7$+;H$o$J$$!%$+$i$$$i$J$$!%$H;W$&!%(B
		 * exception_table $B$O!D(B?
		 */
		for ( i=0; i< (length-(8+code_length)-2); i++ )
		   read_u1(buf); 

	} else {
		int i;
		for ( i=0; i<length; i++ ) {
			read_u1(buf);
		}
	}	
}

void load_pool( struct const_pool *ptr, int count, READ_BUF *buf )
{
	int i;
	for ( i=1; i<count; i++ ) {
		u1 tag = read_u1(buf);
		ptr[i].type = tag;
		
		switch ( tag ) {
		case tag_UTF8: {
			char *str = utf8(buf);
			ptr[i].val1 = (u4)str;
			break;
		}

		case tag_INT:
			ptr[i].val1 = read_u4(buf);
			break;
			
		case tag_CLASS:
		case tag_STRING:
			ptr[i].val1 = read_u2(buf);
			break;

		case tag_FIELDREF: /* val1 ->class | val2 ->name */
		case tag_METHODREF:
		case tag_INTFCMETHREF:
		case tag_NAMEANDTYPE: /* val1 ->name | val2 ->type */
			ptr[i].val1 = read_u2(buf);
			ptr[i].val2 = read_u2(buf);
			break;

		default:
			fetal_error("unrecognize constant pool tag.");
			break;
		}
	}
}


struct first_path
{
	int smax_argnum;	/* $B0lHV0z?t$N?t$,B?$$%a%=%C%I$N0z?t$N?t(B */
	int imax_argnum;
	
	int methodnum;		/* instance,static$B9g$o$;$?%a%=%C%I$N?t(B */

	int static_fieldnum;	
	int instance_fieldnum;
};

/* $B$$$i$J$$(Battribute$B$O<N$F$k(B */
static void discard_attribute( READ_BUF *buf )
{
	int length;
	int i;
	read_u2(buf);
	length = read_u4(buf);

	for ( i=0; i<length; i++ )
		read_u1(buf);
}
		
/* $B>e$N9=B$BN;2>H!%%a%=%C%I$H$+$N?t$r?t$($k(B */
static void first_path( struct first_path *fp, READ_BUF *buf, struct jclass *cls )
{
	unsigned long pos = buf_tell(buf);

	int ifnum=0, sfnum=0;
	int smaxarg=0, imaxarg=0;

	int field_cnt = read_u2(buf);
	int meth_cnt;
	int i;

	for ( i=0; i<field_cnt; i++ ) {
		int attcnt;
		int j;
		int access = read_u2(buf);
		if ( access&ACC_STATIC )
			sfnum++;
		else
			ifnum++;
		read_u4(buf);
		attcnt = read_u2(buf);
		for ( j=0; j<attcnt; j++ ) 
			discard_attribute(buf);
	}

	meth_cnt = read_u2(buf);

	for ( i=0; i<meth_cnt; i++ ) {
		int attcnt;
		int j;
		int descidx;
		int argnum;
		int acc = read_u2(buf);

		read_u2(buf);
		descidx = read_u2(buf);
		argnum = compute_argnum( STRING_CONST(cls,descidx) );

		if ( acc&ACC_STATIC ) {
			if ( argnum > smaxarg )
				smaxarg = argnum;
		} else {
			if ( argnum > imaxarg )
				imaxarg = argnum;
		}
		
		attcnt = read_u2(buf);
		for ( j=0; j<attcnt; j++ )
			discard_attribute(buf);
	}

	buf_setpos(buf,pos);
	fp->smax_argnum=smaxarg;
	fp->imax_argnum=imaxarg;
	fp->methodnum=meth_cnt;
	fp->static_fieldnum=sfnum;
	fp->instance_fieldnum=ifnum;
}

static void load_field( struct jclass *cls, struct first_path *_1path, READ_BUF *buf )
{
	int field_cnt = read_u2(buf);	/* access_flag... */
	int i;
	int field_offset;
	int static_offset=0;

	struct field_bucket *instance_p;
	struct field_bucket *static_p;

	if ( _1path->static_fieldnum!=0 )
		static_p = cls->static_field = xmalloc(sizeof(struct field_bucket)*_1path->static_fieldnum);
	if ( _1path->instance_fieldnum!=0 )
		instance_p = xmalloc( sizeof(struct field_bucket)*_1path->instance_fieldnum);
	cls->instance_field = instance_p;

	cls->static_field_num = _1path->static_fieldnum;
	cls->instance_field_num = _1path->instance_fieldnum;

	if ( _1path->static_fieldnum != 0 )
		cls->class_field_val = xcalloc( _1path->static_fieldnum, sizeof(jvalue) );
	cls->field_size = cls->parent->field_size+cls->instance_field_num;
	field_offset = cls->parent->field_size;

	for ( i=0; i<field_cnt; i++ ) {
		int access=read_u2(buf);
		int attcnt,j;

		if ( access&ACC_STATIC ) {
			static_p->name=STRING_CONST(cls,read_u2(buf));
			static_p->offset=static_offset++;
			static_p++;
		} else {
			instance_p->name = STRING_CONST(cls,read_u2(buf));
			instance_p->offset = field_offset++;
			instance_p++;
		}
		read_u2(buf);	/* descriptor */
		
		attcnt = read_u2(buf);

		/* constant value $B$O(B... ? */
		for ( j=0; j<attcnt; j++ )
			discard_attribute(buf);
	}


}

static int count_override( struct jclass *cls )
{
	int i;
	int count = 0;
	int offset = cls->parent->vtbl_size;

	for ( i=0; i<cls->instance_max_argnum; i++ ) {

		struct method_list *this_p = cls->instance_method[i];

		for ( ; this_p; this_p=this_p->next ) {
			struct method_list *parent_p = cls->parent->instance_method[i];

			for ( ; parent_p; parent_p=parent_p->next ) {
				if ( (strcmp(parent_p->name,this_p->name)==0) &&
				     (strcmp(parent_p->desc,this_p->desc)==0) ) {
					count++;
					this_p->offset = parent_p->offset;
					break;
				}
			}

			if ( parent_p == NULL )
				this_p->offset = offset++;
		}
	}

	return count;
}

static void make_vtbl( struct jclass *cls )
{
	int override_cnt = count_override(cls);

	int i=0;

	cls->vtbl_size += cls->parent->vtbl_size - override_cnt;
	cls->vtbl = xmalloc( sizeof(struct method_list*) * (cls->vtbl_size) );

	memcpy( cls->vtbl, cls->parent->vtbl, sizeof(struct method_list*)*cls->parent->vtbl_size );

	for ( i=0; i<cls->instance_max_argnum; i++ ) {
		struct method_list *p;
		for ( p=cls->instance_method[i]; p; p=p->next ) {
			cls->vtbl[ p->offset ] = p;
		}
	}

}


/*
 * $B3F0z?t$N?t$4$H$N%a%=%C%I$N%j%9%H$r:n$j!$$=$N=*$o$j$r?F$N%a%=%C%I$K7R$2$F$*$/(B
 */
static void load_method( struct jclass *cls, struct first_path *_1path, READ_BUF *buf )
{
	int count = read_u2(buf);
	int imaxarg, smaxarg;
	int i;

	/* meth_p
	 * $B$$$C$+$$$$$C$+$$(Bmalloc$B$H$+5$;}$A0-$$$N$G!$I,MW$J?t$@$1G[Ns$r:n$C$H$/(B */
	struct method_list *meth_p = xmalloc( sizeof(struct method_list)*count);
	struct method_list ***static_p, ***instance_p; /* $B$o$1$o$+$i$s!%(B */

	/* static */
	smaxarg = (_1path->smax_argnum>cls->parent->static_max_argnum) ?
		_1path->smax_argnum: cls->parent->static_max_argnum;
	cls->static_method = xmalloc( sizeof(struct method_list*)*(smaxarg+1) );
	cls->static_max_argnum = smaxarg;

	static_p = xmalloc( sizeof(struct method_list*)*(smaxarg+1) );
	for ( i=0; i<smaxarg+1; i++ )
		static_p[i]=&cls->static_method[i];

	/* instance */
	imaxarg = (_1path->imax_argnum>cls->parent->instance_max_argnum) ?
		_1path->imax_argnum: cls->parent->instance_max_argnum;
	cls->instance_method = xmalloc( sizeof(struct method_list*)*(imaxarg+1) );
	cls->instance_max_argnum = imaxarg;

	instance_p = xmalloc( sizeof(struct method_list**)*(imaxarg+1) );
	for ( i=0; i<imaxarg+1; i++ )
		instance_p[i]=&cls->instance_method[i];

	for ( i=0; i<count; i++ ) {
		struct method_list *meth = meth_p++;
		int j;
		
		int access=read_u2(buf);
		char *name = STRING_CONST(cls,read_u2(buf));

		char *desc = STRING_CONST(cls,read_u2(buf));
		int argnum=compute_argnum(desc);
		int attcnt;

		meth->is_native=0;

		if ( access&ACC_STATIC ) {
			*(static_p[argnum]) = meth;
			static_p[argnum] = &(meth->next);
		} else {
			*(instance_p[argnum]) = meth;
			instance_p[argnum] = &(meth->next);
			argnum++; /* $B0z?t(B0$B$K%$%s%9%?%s%9$X$N%]%$%s%?(B */
			cls->vtbl_size++;
			meth->offset = -1;
		}
		
		meth->name = name;
		meth->next = NULL;
		meth->desc = desc;
		meth->argnum = argnum;
		meth->ofclass = cls;

		attcnt=read_u2(buf);
		for ( j=0; j<attcnt; j++ )
			read_methatt( buf, meth, cls );
	}

	xfree(static_p);

	make_vtbl( cls );

	/* $B?F%/%i%9$N$K7R$2$F$*$/(B */
	imaxarg=cls->parent->instance_max_argnum;
	for ( i=0; i<imaxarg+1; i++ )
		(*instance_p[i]) = cls->parent->instance_method[i];
	xfree(instance_p);
}

/* jclass $B$r:n$k(B */
struct jclass *load_class( READ_BUF *buf )
{
	int pool_cnt;
	int interface_cnt;

	struct jclass *cls = xmalloc( sizeof(struct jclass) );

	struct first_path _1path;
	
	if ( read_u4(buf) != 0xCAFEBABE ) { /* magic number */
		fetal_error("can't load class file.");
		return NULL;
	}
	
	cls = xmalloc( sizeof(struct jclass) );

	read_u4(buf);		/* major version $B$H(B minor version */

	cls->vtbl_size=0;
	
	pool_cnt = read_u2(buf);	/* constant pool count */
	cls->pool_num = pool_cnt;
	cls->constants = xmalloc( sizeof(struct const_pool)*pool_cnt );
	load_pool( cls->constants, pool_cnt, buf );

	/* general information */
	read_u2(buf);		/* access_flag */

	cls->classname = STRING_CONST(cls,cls->constants[read_u2(buf)].val1);
	register_class( cls );
	cls->parent = get_class( STRING_CONST(cls, cls->constants[read_u2(buf)].val1) );
	if ( cls->parent == NULL ) {
		fetal_error("parent $B$,8+$D$+$i$X$s(B.");
	}

	/* interface */
	interface_cnt = read_u2(buf);
	if ( interface_cnt != 0 ) {
		fetal_error("can't use interface.");
	}

	first_path( &_1path, buf, cls );
	load_field( cls, &_1path, buf );

	load_method( cls, &_1path, buf );

	resolve_constant( cls );

	return cls;

//error:
	// free_class(cls);
	return NULL;
}

struct jclass *load_new_class( const char *name )
{
	char *fname=NULL;
	READ_BUF *fp=NULL;
	struct jclass *ret=NULL;
	fname = xmalloc(strlen(name)+5+1); /* ".java"=5 */
	strcpy( fname, name );

	strcat( fname, ".class" );

	fp = readbuf_open( fname );
	if ( fp == NULL )
		goto error;

	ret = load_class( fp );
	if ( ret == NULL )
		goto error;

	goto exit;

error:
	if ( fname ) xfree(fname);
	if ( ret ) xfree(ret);

exit:
	if ( fp ) readbuf_close(fp);
	return ret;
}

	
